#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include <math.h>


#define  CONSTANT_A  6        // Passed in as -D gcc parameter. 14 55 24 24
#define  CONSTANT_B  2        // Passed in as -D gcc parameter.  2 14 16 16
#define  CONSTANT_C  5        // Passed in as -D gcc parameter.  7 36 37 37
#define  CONSTANT_D  5        // Passed in as -D gcc parameter.  5    29  5


typedef  uint16_t  rngword_t;


// Seed the state array.  Seed MUST not be all zero.
static rngword_t  s[2] = {1, 0};


static inline rngword_t  rotl( rngword_t value, int shift )  {
	return (value << shift) | (value >> (sizeof(rngword_t)*8 - shift));
}


static rngword_t  next_xoro( void )
{
	rngword_t  s0 = s[0];
	rngword_t  s1 = s[1];
	rngword_t  result = rotl( s0 + s1, CONSTANT_D ) + s0;  // output hash

	//Perform the iteration
	s1 ^= s0;
	s[0] = rotl( s0, CONSTANT_A ) ^ s1 ^ (s1 << CONSTANT_B);
	s[1] = rotl( s1, CONSTANT_C );

	return result;
}



int  main(int argc, char* argv[])
{
	uint32_t  pfreq[256], zfreq[256], irng, ipair, next_prn, prev_prn;
	uint8_t   *pairs, ifreq, ztally = 0;
	uint64_t  total = 0;
	FILE  *fh;
	struct timespec  start, stop;
	double  duration;



	// initialize pair array
	pairs = (uint8_t *)calloc( 1UL<<32, sizeof(uint8_t) );    // Zero'd 4 GB memory allocation
	if( pairs == NULL )
	{
		printf( "malloc error\n" );
		exit(1);
	}

	clock_gettime( CLOCK_MONOTONIC, &start );
	printf( "Start at %.3f\n", start.tv_sec + start.tv_nsec/1.0e9 );


	// initialize prev_prn
	prev_prn = next_xoro();

	// iterate full period and increment pair values
	for( irng = 0; irng < (1UL<<32)-1; irng++ )
	{
		next_prn = next_xoro();
		ipair = (next_prn << 16) | prev_prn;
//		ipair = (prev_prn << 16) | next_prn;
		pairs[ipair]++;
		prev_prn = next_prn;
	}

	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "Paired at %.3f  -  Duration = %.3f\n", stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	// initialize freq array
	for( irng = 0; irng < 256; irng++ )  pfreq[irng] = zfreq[irng] = 0;

	// read pair array and increment freq values
	irng = 0;
	do {
		ifreq = pairs[irng++];
		pfreq[ifreq]++;
		if( ifreq != 0 )
		{
			zfreq[ztally]++;
			ztally = 0;
		}
		else  ztally++;
	} while( irng );


	clock_gettime( CLOCK_MONOTONIC, &stop );
	duration = (time_t)(stop.tv_sec - start.tv_sec) + (long)(stop.tv_nsec - start.tv_nsec)/1.0e9;
	printf( "Finish at %.3f  -  Total duration = %.3f\n", stop.tv_sec + stop.tv_nsec/1.0e9, duration );


	// write freq array to file
	fh = fopen( "pdist_report.bin", "a" );
	if( fh != NULL )
	{
		fwrite( pfreq, sizeof(pfreq), 1, fh );
		fclose( fh );
	}
	else	printf( "pbin file error\n" );

	fh = fopen( "zdist_report.bin", "a" );
	if( fh != NULL )
	{
		fwrite( zfreq, sizeof(zfreq), 1, fh );
		fclose( fh );
	}
	else	printf( "zbin file error\n" );

	fh = fopen( "pdist_report.txt", "a" );
	if( fh != NULL )
	{
		fprintf( fh, "Frequency distribution of pairs for Xoroshiro32++ [%2d %2d %2d %2d]\n", CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D );

		for( irng = 0; irng < 256; irng++ )
		{
			total += pfreq[irng];
			fprintf( fh, "%3d  %u\n", irng, pfreq[irng] );
		}

		fprintf( fh, "Total = %lu or 0x%lx\n\n", total, total );
		fclose( fh );
	}
	else	printf( "ptxt file error\n" );

	fh = fopen( "zdist_report.txt", "a" );
	if( fh != NULL )
	{
		total = 0;
		fprintf( fh, "Frequency distribution of zero runs for Xoroshiro32++ [%2d %2d %2d %2d]\n", CONSTANT_A, CONSTANT_B, CONSTANT_C, CONSTANT_D );

		for( irng = 0; irng < 256; irng++ )
		{
			total += zfreq[irng];
			fprintf( fh, "%3d  %u\n", irng, zfreq[irng] );
		}

		fprintf( fh, "Total = %lu or 0x%lx\n\n", total, total );
		fclose( fh );
	}
	else	printf( "ztxt file error\n" );


	free( pairs );
	return 0;
}

